#include "nodes/stream/streamNode.hpp"
#include "common/utils.hpp" // Assuming Timer might be here, if used
#include <iostream>        // For std::cerr
#include <thread>          // For std::this_thread::sleep_for
#include <chrono>          // For std::chrono::milliseconds

namespace GNode
{

// --- Private Helper Methods ---

void StreamNode::close_stream() {
    printf("StreamNode [%s]: Closing stream...\n", name_.c_str());
    // Reset pointers, which will call destructors if they are unique owners
    cap_.reset();
    decoder_.reset(); // Decoder depends on demuxer info, close it first potentially
    demuxer_.reset();
    status_ = StreamStatus::CLOSED;
    frame_count_ = -1; // Reset frame count for the next potential stream
}

bool StreamNode::open_stream() {
    // Ensure any previous stream is closed before opening a new one
    close_stream();

    printf("StreamNode [%s]: Attempting to open stream: %s\n", name_.c_str(), stream_url_.c_str());
    status_ = StreamStatus::CLOSED; // Start as closed before trying

    if (decode_type_ == DecodeType::GPU)
    {
        demuxer_ = FFHDDemuxer::create_ffmpeg_demuxer(stream_url_);
        if (demuxer_ == nullptr)
        {
            fprintf(stderr, "StreamNode [%s] Error: GPU demuxer creation failed for %s\n", name_.c_str(), stream_url_.c_str());
            status_ = StreamStatus::OPEN_FAILED;
            return false;
        }

        auto codec_id = demuxer_->get_video_codec();

        decoder_ = FFHDDecoder::create_cuvid_decoder(
            false, FFHDDecoder::ffmpeg2NvCodecId(codec_id), -1, gpu_id_, nullptr, nullptr, true
        );

        if (decoder_ == nullptr)
        {
            fprintf(stderr, "StreamNode [%s] Error: GPU decoder creation failed for %s (Codec: %d)\n", name_.c_str(), stream_url_.c_str(), codec_id);
            demuxer_.reset(); // Clean up demuxer if decoder fails
            status_ = StreamStatus::OPEN_FAILED;
            return false;
        }
        printf("StreamNode [%s]: GPU Demuxer and Decoder created successfully.\n", name_.c_str());
        status_ = StreamStatus::OPENED;
    }
    else // DecodeType::CPU
    {
        cap_ = std::make_shared<cv::VideoCapture>();
        // Optionally set backend preference if needed
        // cap_->open(stream_url_, cv::CAP_FFMPEG);
        if (!cap_->open(stream_url_)) // Check return value of open
        {
            fprintf(stderr, "StreamNode [%s] Error: CPU cv::VideoCapture failed to open %s\n", name_.c_str(), stream_url_.c_str());
            cap_.reset(); // Release the failed object
            status_ = StreamStatus::OPEN_FAILED;
            return false;
        }
        if (!cap_->isOpened()) // Double check
        {
             fprintf(stderr, "StreamNode [%s] Error: CPU cv::VideoCapture not opened after call for %s\n", name_.c_str(), stream_url_.c_str());
             cap_.reset();
             status_ = StreamStatus::OPEN_FAILED;
             return false;
        }
        printf("StreamNode [%s]: CPU cv::VideoCapture opened successfully.\n", name_.c_str());
        status_ = StreamStatus::OPENED;
    }

    frame_count_ = -1; // Reset frame count upon successful open
    return true;
}


// --- Main Work Loop ---

void StreamNode::work()
{
    printf("StreamNode [%s] starting work loop. Decode type: %s\n",
           name_.c_str(), (decode_type_ == DecodeType::GPU ? "GPU" : "CPU"));
    while (running_) // Main loop continues as long as the node is supposed to run
    {
        if (status_ != StreamStatus::OPENED) // Check if stream needs opening/reopening
        {
            printf("StreamNode [%s]: Stream not open (Status: %d). Attempting to open...\n",
                   name_.c_str(), static_cast<int>(status_));

            if (open_stream()) // Try to open
            {
                printf("StreamNode [%s]: Stream opened successfully.\n", name_.c_str());
                // Continue to processing immediately after successful open
            }
            else
            {
                // Opening failed, wait before retrying
                fprintf(stderr, "StreamNode [%s]: Failed to open stream. Retrying in %d ms...\n",
                        name_.c_str(), retry_delay_ms_);
                status_ = StreamStatus::OPEN_FAILED; // Ensure status reflects failure

                // Wait for the specified delay, but check running_ periodically
                auto wakeUpTime = std::chrono::steady_clock::now() + std::chrono::milliseconds(retry_delay_ms_);
                while (running_ && std::chrono::steady_clock::now() < wakeUpTime) {
                     std::this_thread::sleep_for(std::chrono::milliseconds(100)); // Sleep in smaller chunks
                }

                if (!running_) break; // Exit outer loop if stopped during wait
                continue; // Go back to the start of the outer loop to retry opening
            }
        }

        // If we reach here, the stream should be OPENED
        if (status_ == StreamStatus::OPENED)
        {
            printf("StreamNode [%s]: Starting stream processing...\n", name_.c_str());
            if (decode_type_ == DecodeType::CPU)
            {
                process_stream_cpu();
            }
            else
            {
                process_stream_gpu();
            }
            // After processing function returns, the stream might be closed or encountered an error.
            // The loop will re-evaluate the status_ at the beginning.
             printf("StreamNode [%s]: Stream processing finished or stopped (Status: %d).\n",
                   name_.c_str(), static_cast<int>(status_));

            if (status_ == StreamStatus::CLOSED || status_ == StreamStatus::ERROR) 
            {
                 close_stream(); // Ensure resources are released if processing stopped abnormally
                 printf("StreamNode [%s]: Stream closed or errored. Will attempt reconnection if running.\n", name_.c_str());
                  // Optional short delay even after normal close before retry?
                  // std::this_thread::sleep_for(std::chrono::milliseconds(100));
            }
        }
        else 
        {
            // Should not happen if open_stream logic is correct, but good for debugging
            fprintf(stderr, "StreamNode [%s]: Unexpected status %d in work loop.\n", name_.c_str(), static_cast<int>(status_));
             std::this_thread::sleep_for(std::chrono::milliseconds(1000)); // Avoid tight loop on unexpected state
        }

    } // End while(running_)

    printf("StreamNode [%s] work loop finished.\n", name_.c_str());
    close_stream(); // Final cleanup
}


// --- Stream Processing Methods ---

void StreamNode::process_stream_cpu()
{
    if (!cap_ || !cap_->isOpened()) {
        fprintf(stderr, "StreamNode [%s] Error: process_stream_cpu called with closed/invalid VideoCapture.\n", name_.c_str());
        status_ = StreamStatus::ERROR; // Indicate an unexpected state
        return;
    }

    printf("StreamNode [%s]: Processing CPU stream...\n", name_.c_str());
    while (running_ && status_ == StreamStatus::OPENED)
    {
        cv::Mat frame;
        bool success = false;
        try {
             success = cap_->read(frame);
        } catch (const cv::Exception& ex) {
            fprintf(stderr, "StreamNode [%s] Error: Exception during cv::VideoCapture::read(): %s\n", name_.c_str(), ex.what());
            status_ = StreamStatus::ERROR; // Treat OpenCV exception as an error
            break; // Exit processing loop
        }

        if (!success || frame.empty())
        {
            fprintf(stderr, "StreamNode [%s]: Cannot read frame (End of stream or error).\n", name_.c_str());
            status_ = StreamStatus::CLOSED; // Assume normal closure or recoverable error
            break; // Exit processing loop, work() will handle retry/stop
        }

        frame_count_++;
        if (frame_count_ % skip_frame_ != 0)
        {
            continue; // Skip frame
        }

        auto metaData = std::make_shared<meta::MetaData>();
        // Use clone() to ensure the pushed data is independent if frame is reused by VideoCapture
        metaData->image = frame.clone();
        metaData->from = name_;
        // metaData->timestamp = getCurrentTimestamp(); // Add timestamp if useful

        bool pushed = false;
        for (auto& output_buffer : output_buffers_)
        {
            if (output_buffer.second) { // Check if buffer pointer is valid
                 output_buffer.second->push(metaData);
                 pushed = true;
            }
        }
         // Optional: Log if data wasn't pushed anywhere
        // if (!pushed) {
        //     printf("StreamNode [%s]: Warning - Frame %d processed but not pushed to any output buffer.\n", name_.c_str(), frame_count_);
        // }
    }
     printf("StreamNode [%s]: Exiting CPU processing loop (Running: %s, Status: %d).\n",
            name_.c_str(), running_ ? "true" : "false", static_cast<int>(status_));
}

void StreamNode::process_stream_gpu()
{
     if (!demuxer_ || !decoder_) {
        fprintf(stderr, "StreamNode [%s] Error: process_stream_gpu called with invalid demuxer/decoder.\n", name_.c_str());
        status_ = StreamStatus::ERROR;
        return;
    }

    printf("StreamNode [%s]: Processing GPU stream...\n", name_.c_str());
    uint8_t* packet_data = nullptr;
    int packet_size = 0;
    int64_t pts = 0;

    // Send extradata once (important for some codecs)
    demuxer_->get_extra_data(&packet_data, &packet_size)
    if (packet_size > 0) {
        printf("StreamNode [%s]: Sending %d bytes of extradata to decoder.\n", name_.c_str(), packet_size);
        decoder_->decode(packet_data, packet_size);
    } else {
         printf("StreamNode [%s]: No extradata found or needed.\n", name_.c_str());
    }


    unsigned int frame_index = 0; // Consider if this should be member if state needs preserving across reconnects
    while(running_ && status_ == StreamStatus::OPENED)
    {
        // Demux next packet
        bool demux_ok = false;
        try {
             demux_ok = demuxer_->demux(&packet_data, &packet_size, &pts);
        } catch (const std::exception& ex) { // Catch potential exceptions from demuxer implementation
             fprintf(stderr, "StreamNode [%s] Error: Exception during demuxer_->demux(): %s\n", name_.c_str(), ex.what());
             status_ = StreamStatus::ERROR;
             break;
        }

        if (!demux_ok || packet_size <= 0 || !running_) // Check running_ again after potentially blocking demux call
        {
            printf("StreamNode [%s]: Demuxing finished or failed (packet_size: %d, running: %s).\n",
                   name_.c_str(), packet_size, running_ ? "true":"false");
            status_ = StreamStatus::CLOSED; // Assume normal end or recoverable error
            break; // Exit processing loop
        }

        // Decode the packet
        int ndecoded_frame = 0;
        try 
        {
            ndecoded_frame = decoder_->decode(packet_data, packet_size, pts);
        } 
        catch (const std::exception& ex) 
        {
             fprintf(stderr, "StreamNode [%s] Error: Exception during decoder_->decode(): %s\n", name_.c_str(), ex.what());
             status_ = StreamStatus::ERROR;
             break;
         }

        if (ndecoded_frame < 0) 
        {
            fprintf(stderr, "StreamNode [%s] Error: Decoder returned error (%d).\n", name_.c_str(), ndecoded_frame);
            status_ = StreamStatus::ERROR; // Treat decoder error as critical
            break; // Exit processing loop
        }

        // Process decoded frames
        for(int i = 0; i < ndecoded_frame; ++i)
        {
            // Timer timer("StreamNode GPU Frame"); // If using Timer utility
            uint8_t* frame_data = nullptr;
            int64_t frame_pts = 0;

            try 
            {
                 // Pass pointers to get the actual index and PTS for the current frame
                frame_data = decoder_->get_frame(&frame_pts, &frame_index);
            } 
            catch (const std::exception& ex) { // Catch potential exceptions from get_frame implementation
                 fprintf(stderr, "StreamNode [%s] Error: Exception during decoder_->get_frame(): %s\n", name_.c_str(), ex.what());
                 status_ = StreamStatus::ERROR;
                 ndecoded_frame = 0; // Stop processing frames from this packet
                 break; // Break inner frame loop
            }

            if (!frame_data) 
            {
                fprintf(stderr, "StreamNode [%s] Error: Decoder returned null frame data for frame %d.\n", name_.c_str(), i);
                status_ = StreamStatus::ERROR; // Treat null frame data as error
                ndecoded_frame = 0; // Stop processing frames from this packet
                break; // Break inner frame loop
            }


            // Update overall frame count and check skip logic
            frame_count_++;
            if (frame_count_ % skip_frame_ != 0)
            {
                continue; // Skip this decoded frame
            }

            // Important: Create cv::Mat header WITHOUT copying data yet.
            // The data pointer (frame_data) is owned by the decoder.
            // We need to clone it if we push it asynchronously.
            cv::Mat frame_gpu(decoder_->get_height(), decoder_->get_width(), CV_8UC3, frame_data);

            // Create metadata and copy the frame data
            auto metaData = std::make_shared<meta::MetaData>();
            metaData->image = frame_gpu.clone(); // CLONE is crucial here!
            metaData->from = name_;
            // metaData->timestamp = getCurrentTimestamp(); // Or use frame_pts if preferred
            // metaData->frame_idx = frame_count_; // Add frame index if needed downstream

            bool pushed = false;
            for (auto& output_buffer : output_buffers_)
            {
                if (output_buffer.second) 
                {
                    output_buffer.second->push(metaData);
                    pushed = true;
                }
            }
             // Optional: Log if data wasn't pushed anywhere
            // if (!pushed) {
            //     printf("StreamNode [%s]: Warning - GPU Frame %d processed but not pushed to any output buffer.\n", name_.c_str(), frame_count_);
            // }

            // Optional delay if needed for rate limiting, but usually not here
            // std::this_thread::sleep_for(std::chrono::milliseconds(1));
        } // End loop over decoded frames

        // If status changed to ERROR within the frame loop, break the outer while loop
        if (status_ == StreamStatus::ERROR) 
        {
             break;
        }

    }; // End while(running_ && status_ == StreamStatus::OPENED)

	printf("StreamNode [%s]: Exiting GPU processing loop (Running: %s, Status: %d, Total frames processed this session: %d).\n",
           name_.c_str(), running_ ? "true" : "false", static_cast<int>(status_), frame_count_ + 1);
}


}   // namespace GNode